pub mod primality_tests {
    use num::{BigUint, Integer, One, Zero};
    use rand::RngCore;

    fn generate_random_biguint(num_bits: usize) -> BigUint {
        let mut rng = rand::thread_rng();
        let bytes = num_bits / 8 + 1;
        let mut buf = vec![0u8; bytes];
        rng.fill_bytes(&mut buf);
        BigUint::from_bytes_be(&buf)
    }

    
    pub fn is_probably_prime(number: &BigUint, iterations: u32) -> bool {
        if number <= &BigUint::one() || number == &BigUint::from(4u32) {
            return false;
        } else if number <= &BigUint::from(3u32) {
            return true;
        }
        let one = BigUint::one();
        let mut d = number - &one;
        while d.is_even() {
            d /= 2u32;
        }
        for _ in 0..iterations {
            let a = generate_random_biguint((number.bits()-1) as usize);
            let mut x = mod_exp(a.clone(), &d, number);
            if x == one || x == number - &one {
                continue;
            }
            let mut continue_loop = false;
            for _ in 0..(number.bits() - 1) {
                x = mod_exp(x.clone(), &BigUint::from(2u32), number);
                if x == number - &one {
                    continue_loop = true;
                    break;
                }
            }
            if !continue_loop {
                return false;
            }
        }
        true
    }
    
    fn mod_exp(mut base: BigUint, ex: &BigUint, modulus: &BigUint) -> BigUint {
        let mut exp = ex.clone();
        let mut result = BigUint::one();
        base %= modulus;
        while !exp.is_zero() {
            if exp.is_odd() {
                result = (&result * &base) % modulus;
            }
            base = base.pow(2) % modulus;
            exp /= 2u32;
        }
        result
    }
    
}